@@ -1,11 +1,14 @@
 import asyncio
 import contextlib
 import os
+import sys
+
 import pathlib
 import urllib.parse
 from concurrent.futures import ThreadPoolExecutor
 from functools import partial
 
+import parfive
 import aiohttp
 from tqdm import tqdm, tqdm_notebook
 
@@ -55,10 +58,15 @@ class Downloader:
         returned to the existing file, if `True` the file will be downloaded
         and the existing file will be overwritten, if `'unique'` the filename
         will be modified to be unique.
+
+    headers : `dict`
+       Request headers to be passed to the server.
+       Adds `User-Agent` information about `parfive`, `aiohttp` and `python` if not passed explicitely.
+
     """
 
     def __init__(self, max_conn=5, progress=True, file_progress=True,
-                 loop=None, notebook=None, overwrite=False):
+                 loop=None, notebook=None, overwrite=False, headers=None):
 
         self.max_conn = max_conn
         self._start_loop(loop)
@@ -72,6 +80,10 @@ class Downloader:
 
         self.overwrite = overwrite
 
+        self.headers = headers
+        if headers is None or 'User-Agent' not in headers:
+            self.headers = {'User-Agent': f"parfive/{parfive.__version__} aiohttp/{aiohttp.__version__} python/{sys.version[:5]}"}
+
     def _start_loop(self, loop):
         # Setup asyncio loops
         if not loop:
@@ -82,6 +94,8 @@ class Downloader:
             self.loop = loop
             self.run_until_complete = self.loop.run_until_complete
 
+        asyncio.set_event_loop(self.loop)
+
         # Setup queues
         self.http_queue = asyncio.Queue(loop=self.loop)
         self.http_tokens = asyncio.Queue(maxsize=self.max_conn, loop=self.loop)
@@ -130,12 +144,18 @@ class Downloader:
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`
             or `aioftp.ClientSession` depending on the protocol.
+
+        Notes
+        -----
+        Proxy URL is read from the environment variables `HTTP_PROXY` or
+        `HTTPS_PROXY`, depending on the protocol of the `url` passed. Proxy
+        Authentication `proxy_auth` should be passed as a `aiohttp.BasicAuth`
+        object. Proxy Headers `proxy_headers` should be passed as `dict`
+        object.
+
         """
         overwrite = overwrite or self.overwrite
 
-        if 'max_splits' in kwargs:
-            kwargs.pop('max_splits') # popping v1.1 specific kwargs
-
         if path is None and filename is None:
             raise ValueError("Either path or filename must be specified.")
         elif path is None:
@@ -284,12 +304,12 @@ class Downloader:
         return asyncio.gather(*done, return_exceptions=True)
 
     async def _run_http_download(self, main_pb, timeouts):
-        async with aiohttp.ClientSession(loop=self.loop) as session:
+        async with aiohttp.ClientSession(headers=self.headers) as session:
             futures = await self._run_from_queue(self.http_queue, self.http_tokens,
                                                  main_pb, session=session, timeouts=timeouts)
 
             # Wait for all the coroutines to finish
-            done, _ = await asyncio.wait(futures)
+            done, _ = await asyncio.wait(futures, loop=self.loop)
 
             return done
 
@@ -297,7 +317,7 @@ class Downloader:
         futures = await self._run_from_queue(self.ftp_queue, self.ftp_tokens,
                                              main_pb, timeouts=timeouts)
         # Wait for all the coroutines to finish
-        done, _ = await asyncio.wait(futures)
+        done, _ = await asyncio.wait(futures, loop=self.loop)
 
         return done
 
@@ -308,7 +328,8 @@ class Downloader:
             token = await tokens.get()
             file_pb = self.tqdm if self.file_progress else False
             future = asyncio.ensure_future(get_file(session, token=token,
-                                                    file_pb=file_pb, timeouts=timeouts))
+                                                    file_pb=file_pb, timeouts=timeouts),
+                                           loop=self.loop)
 
             def callback(token, future, main_pb):
                 tokens.put_nowait(token)
@@ -321,9 +342,8 @@ class Downloader:
 
         return futures
 
-    @staticmethod
-    async def _get_http(session, *, url, filepath_partial, chunksize=100,
-                        file_pb=None, token, overwrite, timeouts, **kwargs):
+    async def _get_http(self, session, *, url, filepath_partial, chunksize=100,
+                        file_pb=None, token, overwrite, timeouts, max_splits=5, **kwargs):
         """
         Read the file from the given url into the filename given by ``filepath_partial``.
 
@@ -349,6 +369,9 @@ class Downloader:
         token : `parfive.downloader.Token`
             A token for this download slot.
 
+        max_splits: `int`
+            Number of maximum concurrent connections per file.
+
         kwargs : `dict`
             Extra keyword arguments are passed to `aiohttp.ClientSession.get`.
 
@@ -361,6 +384,12 @@ class Downloader:
         """
         timeout = aiohttp.ClientTimeout(**timeouts)
         try:
+            scheme = urllib.parse.urlparse(url).scheme
+            if 'HTTP_PROXY' in os.environ and scheme == 'http':
+                kwargs['proxy'] = os.environ['HTTP_PROXY']
+            elif 'HTTPS_PROXY' in os.environ and scheme == 'https':
+                kwargs['proxy'] = os.environ['HTTPS_PROXY']
+
             async with session.get(url, timeout=timeout, **kwargs) as resp:
                 if resp.status != 200:
                     raise FailedDownload(filepath_partial, url, resp)
@@ -374,28 +403,128 @@ class Downloader:
                                           total=get_http_size(resp))
                     else:
                         file_pb = None
-                    with open(str(filepath), 'wb') as fd:
-                        while True:
-                            chunk = await resp.content.read(chunksize)
-                            if not chunk:
-                                # Close the file progressbar
-                                if file_pb is not None:
-                                    file_pb.close()
 
-                                return str(filepath)
+                    # This queue will contain the downloaded chunks and their offsets
+                    # as tuples: (offset, chunk)
+                    downloaded_chunk_queue = asyncio.Queue()
 
-                            # Write this chunk to the output file.
-                            fd.write(chunk)
+                    download_workers = []
+                    writer = self.loop.create_task(
+                        self._write_worker(downloaded_chunk_queue, file_pb, filepath))
 
-                            # Update the progressbar for file
-                            if file_pb is not None:
-                                file_pb.update(chunksize)
+                    if max_splits and resp.headers.get('Accept-Ranges', None) == "bytes":
+                        content_length = int(resp.headers['Content-length'])
+                        split_length = content_length // max_splits
+                        ranges = [
+                            [start, start + split_length]
+                            for start in range(0, content_length, split_length)
+                        ]
+                        # let the last part download everything
+                        ranges[-1][1] = ''
+                        for _range in ranges:
+                            download_workers.append(
+                                self.loop.create_task(self._http_download_worker(
+                                    session, url, chunksize, _range, timeout, downloaded_chunk_queue, **kwargs
+                                ))
+                            )
+                    else:
+                        download_workers.append(
+                            self.loop.create_task(self._http_download_worker(
+                                session, url, chunksize, None, timeout, downloaded_chunk_queue, **kwargs
+                            ))
+                        )
+
+                    # run all the download workers
+                    await asyncio.gather(*download_workers)
+                    # join() waits till all the items in the queue have been processed
+                    await downloaded_chunk_queue.join()
+                    writer.cancel()
+                    return str(filepath)
 
         except Exception as e:
             raise FailedDownload(filepath_partial, url, e)
 
-    @staticmethod
-    async def _get_ftp(session=None, *, url, filepath_partial,
+    async def _write_worker(self, queue, file_pb, filepath):
+        """
+        Worker for writing the downloaded chunk to the file.
+
+        The downloaded chunk is put into a asyncio Queue by a download worker.
+        This worker gets the chunk from the queue and write it to the file
+        using the specified offset of the chunk.
+
+        Parameters
+        ----------
+
+        queue: `asyncio.Queue`
+             Queue for chunks
+
+        file_pb : `tqdm.tqdm` or `False`
+            Should progress bars be displayed for each file downloaded.
+
+        filepath: `pathlib.Path`
+            Path to the which the file should be downloaded.
+        """
+        with open(filepath, 'wb') as f:
+            while True:
+                offset, chunk = await queue.get()
+
+                f.seek(offset)
+                f.write(chunk)
+                f.flush()
+
+                # Update the progressbar for file
+                if file_pb is not None:
+                    file_pb.update(len(chunk))
+
+                queue.task_done()
+
+    async def _http_download_worker(self, session, url, chunksize, http_range, timeout, queue, **kwargs):
+        """
+        Worker for downloading chunks from http urls.
+
+        This function downloads the chunk from the specified http range and puts the chunk in the
+        asyncio Queue. If no range is specified, then the whole file is downloaded via chunks
+        and put in the queue.
+
+        Parameters
+        ----------
+
+        session : `aiohttp.ClientSession`
+            The `aiohttp.ClientSession` to use to retrieve the files.
+
+        url : `str`
+            The url to retrieve.
+
+        chunksize : `int`
+            The number of bytes to read into the file at a time.
+
+        http_range: (`int`, `int`) or `None`
+            Start and end bytes of the file. In None, then no `Range` header is specified
+            in request and the whole file will be downloaded.
+
+        queue: `asyncio.Queue`
+             Queue to put the download chunks.
+
+        kwargs : `dict`
+            Extra keyword arguments are passed to `aiohttp.ClientSession.get`.
+        """
+        headers = kwargs.pop('headers', {})
+        if http_range:
+            headers['Range'] = 'bytes={}-{}'.format(*http_range)
+            # init offset to start of range
+            offset, _ = http_range
+        else:
+            offset = 0
+
+        async with session.get(url, timeout=timeout, headers=headers, **kwargs) as resp:
+            while True:
+                chunk = await resp.content.read(chunksize)
+                if not chunk:
+                    break
+                await queue.put((offset, chunk))
+                offset += len(chunk)
+
+    async def _get_ftp(self, session=None, *, url, filepath_partial,
                        file_pb=None, token, overwrite, timeouts, **kwargs):
         """
         Read the file from the given url into the filename given by ``filepath_partial``.
@@ -433,7 +562,7 @@ class Downloader:
         try:
             async with aioftp.ClientSession(parse.hostname, **kwargs) as client:
                 if parse.username and parse.password:
-                    client.login(parse.username, parse.password)
+                    await client.login(parse.username, parse.password)
 
                 # This has to be done before we start streaming the file:
                 total_size = await get_ftp_size(client, parse.path)
@@ -447,20 +576,38 @@ class Downloader:
                     else:
                         file_pb = None
 
-                    with open(str(filepath), 'wb') as fd:
-                        async for chunk in stream.iter_by_block():
-                            # Write this chunk to the output file.
-                            fd.write(chunk)
+                    downloaded_chunks_queue = asyncio.Queue()
+                    download_workers = []
+                    writer = self.loop.create_task(
+                        self._write_worker(downloaded_chunks_queue, file_pb, filepath))
 
-                            # Update the progressbar for file
-                            if file_pb is not None:
-                                file_pb.update(len(chunk))
+                    download_workers.append(
+                        self.loop.create_task(self._ftp_download_worker(stream, downloaded_chunks_queue))
+                    )
 
-                        # Close the file progressbar
-                        if file_pb is not None:
-                            file_pb.close()
+                    await asyncio.gather(*download_workers)
+                    await downloaded_chunks_queue.join()
+                    writer.cancel()
 
-                        return str(filepath)
+                    return str(filepath)
 
         except Exception as e:
             raise FailedDownload(filepath_partial, url, e)
+
+    async def _ftp_download_worker(self, stream, queue):
+        """
+        Similar to `Downloader._http_download_worker`. See that function's documentation for more info.
+
+        Parameters
+        ----------
+
+        stream: `aioftp.StreamIO`
+            Stream of the file to be downloaded.
+        queue: `asyncio.Queue`
+             Queue to put the download chunks.
+        """
+        offset = 0
+        async for chunk in stream.iter_by_block():
+            # Write this chunk to the output file.
+            await queue.put((offset, chunk))
+            offset += len(chunk)
